/*  Universal Vehicle Router (UVR)
    Robert Harder
    rharder@usa.net
    
    This code is public domain. While you are free to use this
    code however you like, I would appreciate a brief mention
    if you used it and a pointer to its web site:
        http://www.crosswinds.net/~rharder/uvr
 */
package mil.af.afit.uvr;


/**
 * <code>HSolver</code>s take the information given to the <code>HEngine</code>
 * and try to solve the problem of servicing orders with vehicles. Solvers have
 * access to a list of good starting solutions generated by a heuristic developed
 * by Gary Kinney at the Air Force Institute of Technology (www.afit.af.mil).
 *
 * @author Robert Harder, rharder@usa.net
 * @since 1.0
 * @see HOrder
 * @see HVehicle
 */
public abstract class HSolver
{
    /**
     * Name of the solver
     *
     * @return name of the solver
     * @since 1.0
     */
    public abstract String getName();
    
    
    /**
     * Short description of the solver. May be used for tooltips
     *
     * @return short decription of the solver
     * @since 1.0
     */
    public abstract String getShortDescription();
    
    
    /**
     * Long description of the solver. 
     *
     * @return long decription of the solver
     * @since 1.0
     */
    public abstract String getLongDescription();
    
    
    /**
     * Credits for the solver (authors, etc).
     *
     * @return credits for the solver
     * @since 1.0
     */
    public abstract String getCredits();
    
    
    /**
     * Version of the solver
     *
     * @return version of the solver
     * @since 1.0
     */
    public abstract String getVersion();
    
    
    /**
     * Stop the solver as soon as possible. Be sure to call 
     * <code>fireSolverFinishedEvent</code>.
     *
     * @since 1.0
     * @see HSolverEvent
     */
    public abstract void stopSolving();
    
    
    /**
     * If true, the <code>HEngine</code> will generate a list of good
     * starting solutions based on a heuristic developed by Gary Kinney
     * for the Air Force Institute of Technology
     * (<A HREF="http://www.afit.af.mil" TARGET="_parent">www.afit.af.mil</A>).
     *
     * @return whether or not the solver would like starting solutions
     * @since 1.0
     * @see HEngine
     */
    public abstract boolean requestsStartingSolutions();
    
    /**
     * Called when the <code>HEngine</code> has started solving. The
     * <var>solutions</var> array will either be the starting solutions,
     * if requested, or a single solution with all orders in the dummy tour.
     *
     * @param solutions starting solution(s)
     * @param evaluator evaluator to use with the solutions
     * @param requestedSolutionEffort effort requested by the user in the range of zero to one
     * @param numberOfThreads suggested number of threads to use
     * @since 1.0
     * @see HSolution
     * @see HEvaluator
     */
    public abstract void initializeSolver( HSolution[] solutions,
        HEvaluator evaluator, double requestedSolutionEffort, 
        int numberOfThreads );
    
    /**
     * Called when the <code>HEngine</code> is ready for the solver to start.
     * It is not necessary to return immediate control because the method is
     * called from within a thread dedicated to calling this method.
     *
     * @since 1.0
     */
    public abstract void startSolving();

    /**
     * Return the best solution found so far.
     *
     * @return the best solution
     * @since 1.0
     * @see HSolution
     */
    public abstract HSolution getBestHSolution();
    
    // float used because it is atomic
    private float percentDone = 0;
    
    /**
     * Call to set the percent done for the solver, between zero and one.
     * 
     * @param percentDone percent done, between zero and one
     * @since 1.0
     */
    public final void setPercentDone( float percentDone )
    {   this.percentDone = percentDone;
        fireSolverPercentDoneEvent();
    }   // end setPercentDone
    
    /** 
     * Return the percent done for the solver, between zero and one.
     *
     * @return percent done
     * @since 1.0
     */
    public final float getPercentDone()
    {   return this.percentDone;
    }   // end getPercentDone
    
    
/* ********  E V E N T   C O D E  ******** */

    /**
     * The list of <code>HSolverListener</code>s that have
     * requested to be notified of solver events. Generally
     * this will only contain a reference to the UVR engine.
     * 
     * @since 1.0
     */
        
    private HSolverListener[] HSolverListenerList =
        new HSolverListener[0];
    
    /**
     * Add a listener to the solver. This listener will
     * be notified when solver events happen.
     * 
     * @param listener the listener to notify of solver events
     * @since 1.0
     * @see HSolverListener
     * @see HSolverEvent
     */
    public final synchronized void 
    addHSolverListener( HSolverListener listener )
    {  
        HSolverListener[] list = new HSolverListener[
            HSolverListenerList.length + 1 ];
        for( int i = 0; i < list.length - 1; i++ )
            list[i] = HSolverListenerList[i];
                
        list[ list.length-1 ] = listener;
        HSolverListenerList = list;
    }  // end addHSolverListener
    
    
    /**
     * Remove a listener from the solver. This listener will
     * no longer be notified when solver events happen.
     * 
     * @param listener the listener which will no longer 
     *                 receive solver events
     * @since 1.0
     * @see HSolverListener
     * @see HSolverEvent
     */
    public final synchronized void 
    removeHSolverListener( HSolverListener listener )
    {  // Find location of listener to remove
        int index = -1;
        int j = 0;
        while( index < 0 && j < HSolverListenerList.length )
        {   if( HSolverListenerList[j] == listener )
                index = j;
            else j++;
        }   // end while: through list
            
        // If index is less than zero then it wasn't in the list
        if( index >= 0 )
        {   HSolverListener[] list = new HSolverListener[
                HSolverListenerList.length - 1 ];
            for( int i = 0; i < list.length; i++ )
                list[i] = HSolverListenerList[ i < index ? i : i + 1 ];
            HSolverListenerList = list;
        }   // end if: listener was in the list
    }  // end removeHSolverListener
    
    
    /**
     * Used internally to loop through listener list and send
     * solver events.
     * 
     * @param e the <code>HSolverEvent</code> to send out
     * @since 1.0
     * @see HSolverListener
     * @see HSolverEvent
     */
    private final void fireHSolverEvent( HSolverEvent e )
    {   
        synchronized( this )
        {   for( int i = 0; i < HSolverListenerList.length; i++ )
                HSolverListenerList[i].HSolverEventPerformed( e );
        }   // end synchronized
    }   // end fireHSolverEvent
    
    
    /**
     * This should be called when the solver starts solving.
     * Interested objects will be notified that the solver
     * has started
     * 
     * @since 1.0
     * @see HSolverListener
     * @see HSolverEvent
     */
    public final void fireSolverStartedEvent()
    {   HSolverEvent e = new HSolverEvent(
            HSolverEvent.SOLVER_STARTED, this );
        fireHSolverEvent( e );
    }   // end fireSolverStartedEvent
    
    
    /**
     * This method should be called when the solver has finished
     * working. Interested objects, mainly the UVR engine, will
     * be notified that the solver has finished.
     * 
     * @since 1.0
     * @see HSolverListener
     * @see HSolverEvent
     */
    public final void fireSolverFinishedEvent()
    {   HSolverEvent e = new HSolverEvent( 
            HSolverEvent.SOLVER_FINISHED, this );
        fireHSolverEvent( e );
    }   // end fireSolverFinishedEvent
    
    /**
     * This method should be called periodically to keep the
     * user up to date on the solver's progress.
     * 
     * @param percentDone percent completed
     * @since 1.0
     * @see HSolverListener
     * @see HSolverEvent
     */
    public final void fireSolverPercentDoneEvent()
    {   HSolverEvent e = new HSolverEvent(
            HSolverEvent.PERCENT_DONE, this );
        fireHSolverEvent( e );
    }   // /end fireSolverProgressEvent
    
    
    /**
     * This method should be called when a new best solution is found.
     * 
     * @param percentDone percent completed
     * @since 1.0
     * @see HSolverListener
     * @see HSolverEvent
     */
    public final void fireNewBestSolutionEvent()
    {   HSolverEvent e = new HSolverEvent(
            HSolverEvent.NEW_BEST_SOLUTION, this );
        fireHSolverEvent( e );
    }   // /end fireNewBestSolutionEvent
    
    
    
}   // end abstract class HSolver