/*  Universal Vehicle Router (UVR)
    Robert Harder
    rharder@usa.net
    
    This code is public domain. While you are free to use this
    code however you like, I would appreciate a brief mention
    if you used it and a pointer to its web site:
        http://www.crosswinds.net/~rharder/uvr
 */
/*  This code uses the Java Tabu Search Package written by Robert Harder
    and available at http://www.crosswinds.net/~rharder/tabusearch/
 */
package mil.af.afit.uvr.solvers.tabusearch4uvr;

import mil.af.afit.uvr.*;

/**
 * This solver takes the list of starting solutions from the tour building
 * heuristic and performs a tabu search on each one, starting with the best
 * and ending with either the last or after a certain number of starting
 * solutions have yielded no better answers.
 * <P>
 * A more detailed discussion of the tabu search used here can be found in
 * Robert Harder's thesis, <I>A Java Universal Vehicle Router in Support
 * of Routing Unmanned Aerial Vehicles</I>, from the <A HREF="http://www.afit.af.mil" TARGET="_parent">
 * Air Force Institute of Technology</A>
 * <P>
 * This tabu search relies on the <A HREF="http://www.crosswinds.net/~rharder/tabusearch/" TARGET="_parent">
 * Java Tabu Search Package</A> built by Robert Harder.
 *
 * @author Robert Harder, rharder@usa.net
 * @since 1.0
 * @see mil.af.afit.uvr.HSolver
 */
public class TS4UVRSolver extends HSolver
{
    TS4UVRRunner runner = null;
    HEvaluator evaluator = null;
    private int numberOfThreads = 1;
    private static boolean DEBUGGING = false;
    /**
     * Name of the solver
     */
    public final static String NAME = "Sample Adaptive Tabu Search";
    
    
    /**
     * Long description of the solver
     */
    public final static String LONG_DESCR = 
        "This meta-heuristic takes the starting solutions generated by " +
        "the UVR and searches through them until it determines that there " +
        "is nothing more to be gained by the searching. Each starting " +
        "solution gets a predefined number of iterations which is increased " +
        "if the solution improves as the search goes on. For more information " +
        "see Robert Harder's thesis (2000) from the Air Force Institute of " +
        "Technology (www.afit.af.mil).";
    
    
    /**
     * Short description of the solver
     */
    public final static String SHORT_DESCR = "Searches through a list of starting solutions";
    
    
    
    
    protected final HEvaluator getHEvaluator()
    {   return evaluator;
    }   // end getHEvaluator
    
    public final boolean isDebugging()
    {   return DEBUGGING;
    }   // end isDebugging
    
    public final void setDebugging( boolean state )
    {   DEBUGGING = state;
    }   // end setDebugging
    
/* ********  H S O L V E R   C O D E  ******** */
    
    public String getName()
    {   return NAME;
    }   // end getName
    
    public String getShortDescription()
    {   return SHORT_DESCR;
    }   // end getShortDescription
    
    public String getLongDescription()
    {   return LONG_DESCR;
    }   // end getLongDescription
    
    public String getCredits()
    {   return "Written by Robert Harder (rharder@usa.net)";
    }   // end getCredits
    
    public String getVersion()
    {   return "1.0";
    }   // end getVersion
    
    public boolean requestsStartingSolutions()
    {   return true;
    }   // end requestsStartingSolutions
    
    
    public HSolution getBestHSolution()
    {   //System.out.println( "TS4UVRSolver.getBestHSolution: " + runner.getBestHSolution().isValid() );
        return runner.getBestHSolution();
    }   // end getBestHSolution
    
    
    public void initializeSolver( 
    HSolution[] soln, HEvaluator evaluator,
    double requestedSolutionEffort, int numberOfThreads )
    {
        this.evaluator = evaluator;
        this.numberOfThreads = numberOfThreads;
        
        // Convert solutions to TSSolutions to comply with
        // net.usa.rharder.tabusearch.Solution which TSSolution
        // extends.
        TS4UVRSolution[] solutions = new TS4UVRSolution[ soln.length ];
        for( int i = 0; i < soln.length; i++ )
            solutions[i] = new TS4UVRSolution( soln[i] );
        
        // Make a new runner which does the work
        runner = new TS4UVRRunner( 
            requestedSolutionEffort, solutions, numberOfThreads, this );
        Thread t = new Thread( runner );
        t.setDaemon(true);
        t.start();
    }   // end initializeSolver
    
    public void startSolving()
    {
        // Make sure runner is not null for some reason
        if( runner != null )
        {
            // This method call is required to comply with
            // the HSolver interface.
            fireSolverStartedEvent();

            // Start the runner going which does the real work.
            runner.startSolving();
        }   // end if
    }   // end startSolving
    
    public final void stopSolving()
    {   if( runner != null )
            runner.stopSolving();
    }   // end stopSolving
    
    
/* ********  D E B U G G I N G   C O D E  ******** */
    
    public final static void out( Object obj )
    {   if( DEBUGGING )
            System.out.println( obj );
    }   // end out
    
    public final static void err( Exception e )
    {   if( DEBUGGING )
        {   System.err.println( e );
            e.printStackTrace();
        }   // end if
    }   // end err
    
}   // end TS4UVRSolver